"""
statsdmetrics.client.timing
---------------------------
Easy to use classes to send timing metrics

:license: released under the terms of the MIT license.
For more information see LICENSE or README files, or
https://opensource.org/licenses/MIT.
"""
import functools
from datetime import datetime
from time import time

try:
    from typing import Union, Callable, Any, Tuple, Dict
except ImportError:
    Union, Callable, Any, Tuple, Dict = None, None, None, None, None  # type: ignore


def assert_timestamp(timestamp):
    # type: (Any) -> None
    assert isinstance(timestamp, int) or isinstance(timestamp, float)


def assert_sample_rate(rate):
    # type: (Any) -> None
    assert isinstance(rate, int) or isinstance(rate, float)
    assert 0 <= rate <= 1


class ClientWrapper(object):
    def __init__(self, client):
        self._client = client

    @property
    def client(self):
        return self._client

    @client.setter
    def client(self, client):
        self._client = client


class SampleRateMixIn(object):
    def __init__(self, rate=1):
        # type: (float) -> None
        assert_sample_rate(rate)
        self._rate = rate

    @property
    def rate(self):
        return self._rate

    @rate.setter
    def rate(self, rate):
        # type: (float) -> None
        assert_sample_rate(rate)
        self._rate = rate


class Chronometer(ClientWrapper, SampleRateMixIn):
    def __init__(self, client, rate=1):
        # type: (Any, float) -> None
        SampleRateMixIn.__init__(self, rate)
        ClientWrapper.__init__(self, client)

    def since(self, name, timestamp, rate=None):
        # type (str, Union[float, datetime], float) -> Chronometer
        if rate is None:
            rate = self._rate
        else:
            assert_sample_rate(rate)

        self._client.timing_since(name, timestamp, rate)
        return self

    def time_callable(self, name, target, rate=None, args=(), kwargs={}):
        # type: (str, Callable, float, Tuple, Dict) -> Chronometer
        """Send a Timer metric calculating duration of execution of the provided callable"""
        assert callable(target)
        if rate is None:
            rate = self._rate
        else:
            assert_sample_rate(rate)
        start_time = time()  # type: float
        result = target(*args, **kwargs)
        self.since(name, start_time, rate)
        return result

    def wrap(self, name, rate=None):
        # type: (str, float) -> Callable
        if rate is None:
            rate = self._rate
        else:
            assert_sample_rate(rate)

        def create_decorator(func):
            # type: (Callable) -> Callable
            @functools.wraps(func)
            def decorator(*args, **kwargs):
                return self.time_callable(name, func, rate, args, kwargs)
            return decorator
        return create_decorator


class Stopwatch(ClientWrapper, SampleRateMixIn):
    def __init__(self, client, name, rate=1, reference=None):
        # type: (Any, str, float, float) -> None
        if reference is None:
            reference = time()
        else:
            assert_timestamp(reference)
        SampleRateMixIn.__init__(self, rate)
        ClientWrapper.__init__(self, client)
        self._name = str(name)
        self._paused_duration = 0
        self._reference = reference

    @property
    def reference(self):
        return self._reference

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name):
        # type: (str) -> None
        self._name = str(name)

    def reset(self):
        # type: () -> Stopwatch
        """Reset stop watch by setting now as reference"""
        self._reference = time()
        return self

    def send(self, rate=None):
        # type: (float) -> Stopwatch
        if rate is None:
            rate = self._rate
        else:
            assert_sample_rate(rate)
        self.client.timing_since(self._name, self._reference, rate)
        return self

    def __enter__(self):
        self.reset()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.send()

__all__= ['Chronometer', 'Stopwatch']



"""
tests
-----
statsdmetrics unit tests

:license: released under the terms of the MIT license.
For more information see LICENSE or README files, or
https://opensource.org/licenses/MIT.
"""
import sys
from os.path import dirname
import unittest

try:
    import unittest.mock as mock
except ImportError:
    import mock

project_dir = dirname(dirname(__file__))
if project_dir not in sys.path:
    sys.path.insert(0, project_dir)

from statsdmetrics.client import Client, DEFAULT_PORT


class MockMixIn(object):
    """Base test case to patch socket module for tests"""

    def doMock(self):
        patcher = mock.patch('statsdmetrics.client.socket.gethostbyname')
        self.mock_gethost = patcher.start()
        self.addCleanup(patcher.stop)
        self.mock_gethost.return_value = "127.0.0.2"

        patcher = mock.patch('statsdmetrics.client.random')
        self.mock_random = patcher.start()
        self.addCleanup(patcher.stop)
        self.mock_random.return_value = 0.3

        patcher = mock.patch('statsdmetrics.client.socket.socket')
        self.mock_socket = patcher.start()
        self.addCleanup(patcher.stop)

        self.mock_sendto = mock.MagicMock()
        self.mock_socket.sendto = self.mock_sendto

        self.mock_sendall = mock.MagicMock()
        self.mock_socket.sendall = self.mock_sendall


class ClientTestCaseMixIn(MockMixIn):

    def setUp(self):
        self.doMock()
        self.clientClass = Client

    def test_init_and_properties(self):
        default_client = self.clientClass("127.0.0.1")
        self.assertEqual(default_client.host, "127.0.0.1")
        self.assertEqual(default_client.port, DEFAULT_PORT)
        self.assertEqual(default_client.prefix, "")

        client = self.clientClass("stats.example.org", 8111, "region")
        self.assertEqual(client.host, "stats.example.org")
        self.assertEqual(client.port, 8111)
        self.assertEqual(client.prefix, "region")

    def test_port_number_should_be_valid(self):
        self.assertRaises(AssertionError, self.clientClass, "host", -1)
        self.assertRaises(AssertionError, self.clientClass, "host", 0)
        self.assertRaises(AssertionError, self.clientClass, "host", 65536)

    def test_host_is_readonly(self):
        client = self.clientClass("localhost")
        with self.assertRaises(AttributeError):
            client.host = "example.org"

    def test_port_is_readonly(self):
        client = self.clientClass("localhost")
        with self.assertRaises(AttributeError):
            client.port = 8888

    def test_remote_address_is_readonly(self):
        client = self.clientClass("localhost")
        with self.assertRaises(AttributeError):
            client.remote_address = ("10.10.10.1", 8125)


class BatchClientTestCaseMixIn(ClientTestCaseMixIn):

    def test_init_and_properties(self):
        default_client = self.clientClass("127.0.0.1")
        self.assertEqual(default_client.host, "127.0.0.1")
        self.assertEqual(default_client.port, DEFAULT_PORT)
        self.assertEqual(default_client.prefix, "")
        self.assertGreater(default_client.batch_size, 0)

        client = self.clientClass("stats.example.org", 8111, "region", 1024)
        self.assertEqual(client.host, "stats.example.org")
        self.assertEqual(client.port, 8111)
        self.assertEqual(client.prefix, "region")
        self.assertEqual(client.batch_size, 1024)

    def test_batch_size_should_be_positive_int(self):
        self.assertRaises(
            ValueError, self.clientClass, "localhost", batch_size="not number")
        self.assertRaises(
            AssertionError, self.clientClass, "localhost", batch_size=-1)

    def test_batch_size_is_read_only(self):
        client = self.clientClass("localhost")
        with self.assertRaises(AttributeError):
            client.batch_size = 512


class BaseTestCase(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        unittest.TestCase.__init__(self, *args, **kwargs)
        if not hasattr(self, 'assertRegex'):
            self.assertRegex = self.assertRegexpMatches



"""
tests.test_client_timer
-----------------------
unit tests for timer module

:license: released under the terms of the MIT license.
For more information see LICENSE or README files, or
https://opensource.org/licenses/MIT.
"""
from time import time, sleep
from statsdmetrics.client import Client
from statsdmetrics.client.timing import Chronometer, Stopwatch

try:
    import unittest.mock as mock
except ImportError:
    import mock

from . import BaseTestCase


class TestChronometer(BaseTestCase):
    def setUp(self):
        self.client = Client('127.0.0.1')
        self.request_mock = mock.MagicMock()
        self.client._request = self.request_mock
        self.chronometer = Chronometer(self.client)

    def test_sample_rate_configuration(self):
        self.assertEqual(self.chronometer.rate, 1)
        chronometer = Chronometer(self.client, 0.3)
        self.assertEqual(chronometer.rate, 0.3)
        with self.assertRaises(AssertionError):
            chronometer.rate = "not a number"
        with self.assertRaises(AssertionError):
            chronometer.rate = 2
        with self.assertRaises(AssertionError):
            chronometer.rate = -0.3

    def test_get_set_client(self):
        self.assertEqual(self.chronometer.client, self.client)
        client = Client('127.0.0.2')
        chronometer = Chronometer(client)
        self.assertEqual(chronometer.client, client)
        chronometer.client = self.client
        self.assertEqual(chronometer.client, self.client)

    def test_since(self):
        start_time = time()
        sleep(0.01)
        self.assertEqual(self.chronometer.since("event", start_time), self.chronometer)
        self.assertEqual(self.request_mock.call_count, 1)
        (request,) = self.request_mock.call_args[0]
        self.assertRegex('event:[1-9]\d{0,3}\|ms', request)

        self.request_mock.reset_mock()
        self.assertEqual(self.chronometer.since("event", start_time, rate=0), self.chronometer)
        self.assertEqual(self.request_mock.call_count, 0)

        with self.assertRaises(AssertionError):
            self.chronometer.since("event", start_time, rate=-0.2)
        with self.assertRaises(AssertionError):
            self.chronometer.since("event", start_time, rate=1.01)

    def test_time_callable(self):
        self.assertRaises(AssertionError, self.chronometer.time_callable, "event", "I'm not callable")
    
        self.assertEqual(
            self.chronometer.time_callable("event", self.wait_a_while),
            "waited"
        )
        self.assertEqual(self.request_mock.call_count, 1)
        request_args = self.request_mock.call_args[0]
        self.assertEqual(len(request_args), 1)
        request = request_args[0]
        self.assertRegex(request, "event:[1-9]\d{0,3}\|ms")
    
        self.request_mock.reset_mock()
        self.chronometer.time_callable("low.rate", self.wait_a_while, rate=0.0001)
        self.assertEqual(self.request_mock.call_count, 0)

        args_passed = []
    
        def store_args(*args, **kwargs):
            args_passed.append(args)
            args_passed.append(kwargs)
            sleep(0.01)
    
        self.request_mock.reset_mock()
        self.chronometer.time_callable("with_args", store_args, 1, ("arg1", "arg2"), dict(named_arg="named_value"))
        self.assertEqual(self.request_mock.call_count, 1)
        request_args = self.request_mock.call_args[0]
        self.assertEqual(len(request_args), 1)
        (request,) = request_args
        self.assertRegex(request, "with_args:[1-9]\d{0,3}\|ms")
        self.assertEqual(args_passed, [("arg1", "arg2"), dict(named_arg="named_value")])

        with self.assertRaises(AssertionError):
            self.chronometer.time_callable("event", self.wait_a_while, rate=-0.2)
        with self.assertRaises(AssertionError):
            self.chronometer.time_callable("event", self.wait_a_while, rate=1.01)

    def test_timer_function_decorator(self):
        nap_calls = []

        @self.chronometer.wrap(name="event")
        def take_a_nap():
            nap_calls.append("nap called")
            sleep(0.1)

        take_a_nap()
        self.assertEqual(nap_calls, ["nap called"])
        self.assertEqual(self.request_mock.call_count, 1)
        request_args = self.request_mock.call_args[0]
        self.assertEqual(len(request_args), 1)
        request = request_args[0]
        self.assertRegex(request, "event:[1-9]\d{0,3}\|ms")

        self.request_mock.reset_mock()
        nap_low_rate_calls = []

        @self.chronometer.wrap("event", 0.001)
        def nap_low_rate():
            nap_low_rate_calls.append("low rate called")
            sleep(0.01)

        nap_low_rate()
        self.assertEqual(nap_low_rate_calls, ["low rate called"])
        self.assertEqual(self.request_mock.call_count, 0)

        with self.assertRaises(AssertionError):
            @self.chronometer.wrap("invalid_rate", -0.1)
            def wont_be_called():
                pass

    def wait_a_while(self):
        sleep(0.01)
        return "waited"


class TestStopwatch(BaseTestCase):
    def setUp(self):
        self.client = Client('127.0.0.1')
        self.request_mock = mock.MagicMock()
        self.client._request = self.request_mock
        self.metric_name = "timed_event"
        self.stopwatch = Stopwatch(self.client, self.metric_name)

    def test_name(self):
        stopwatch = Stopwatch(self.client, "new_watch")
        self.assertEqual(stopwatch.name, "new_watch")
        stopwatch.name = self.metric_name
        self.assertEqual(stopwatch.name, self.metric_name)

    def test_sample_rate_configuration(self):
        self.assertEqual(self.stopwatch.rate, 1)
        stopwatch = Stopwatch(self.client, "new_watch", rate=0.3)
        self.assertEqual(stopwatch.rate, 0.3)
        with self.assertRaises(AssertionError):
            stopwatch.rate = "not a number"
        with self.assertRaises(AssertionError):
            stopwatch.rate = 2
        with self.assertRaises(AssertionError):
            stopwatch.rate = -0.3

    def test_get_set_client(self):
        self.assertEqual(self.stopwatch.client, self.client)
        client = Client('127.0.0.2')
        stopwatch = Stopwatch(client, "new_watch")
        self.assertEqual(stopwatch.client, client)
        stopwatch.client = self.client
        self.assertEqual(stopwatch.client, self.client)

    def test_reset(self):
        original_reference = self.stopwatch.reference
        sleep(0.01)
        self.assertEqual(self.stopwatch.reset(), self.stopwatch)
        self.assertGreaterEqual(self.stopwatch.reference, original_reference)

    def test_send(self):
        sleep(0.01)
        self.stopwatch.send()
        self.assertEqual(self.request_mock.call_count, 1)
        request_args = self.request_mock.call_args[0]
        self.assertEqual(len(request_args), 1)
        request = request_args[0]
        self.assertRegex(request, "timed_event:[1-9]\d{0,3}\|ms")

        self.request_mock.reset_mock()
        self.stopwatch.send(rate=0)
        self.assertEqual(self.request_mock.call_count, 0)

    def test_stopwatch_as_context_manager(self):
        original_reference = self.stopwatch.reference
        with self.stopwatch:
            sleep(0.01)
        self.assertGreaterEqual(self.stopwatch.reference, original_reference, "stop watch as context manager resets")
        self.assertEqual(self.request_mock.call_count, 1)
        request_args = self.request_mock.call_args[0]
        self.assertEqual(len(request_args), 1)
        request = request_args[0]
        self.assertRegex(request, "timed_event:[1-9]\d{0,3}\|ms")

        self.request_mock.reset_mock()
        stopwatch = Stopwatch(self.client, "low_rate", rate=0)
        with stopwatch:
            sleep(0.01)
        self.assertEqual(self.request_mock.call_count, 0)



"""
tests.test_metrics
------------------
unit tests for module functions metric classes.

:license: released under the terms of the MIT license.
For more information see LICENSE or README files, or
https://opensource.org/licenses/MIT.
"""

import unittest

from statsdmetrics import (Counter, Timer,
                           Gauge, Set, GaugeDelta,
                           normalize_metric_name,
                           parse_metric_from_request
                           )


class TestMetrics(unittest.TestCase):
    def test_normalize_metric_names_keeps_good_names(self):
        self.assertEqual(
            "metric.good.name",
            normalize_metric_name("metric.good.name")
        )

    def test_normalize_metric_names_replaces_spaces(self):
        self.assertEqual(
            "metric_name_with_spaces",
            normalize_metric_name("metric name with spaces")
        )

    def test_normalize_metric_names_replaces_slashes_and_backslashes(self):
        self.assertEqual(
            "metric-name-with-slashes",
            normalize_metric_name("metric/name\\with/slashes")
        )

    def test_normalize_names_removes_none_alphanumeric_dashes(self):
        self.assertEqual(
            "namewithinvalidcharsandall",
            normalize_metric_name("#+name?with~invalid!chars(and)all*&")
        )

    def test_parse_metric_from_request_requires_string(self):
        self.assertRaises(AssertionError, parse_metric_from_request, 10)
        self.assertRaises(AssertionError, parse_metric_from_request, 2.2)

    def test_parse_counter_metric_from_request(self):
        self.assertEqual(
            Counter("sales", 10),
            parse_metric_from_request("sales:10|c")
        )
        self.assertEqual(
            Counter("with rate?", 0, 1),
            parse_metric_from_request("with rate?:0|c|@1")
        )
        self.assertEqual(
            Counter("accounts.active.disable", 4567, 1.0),
            parse_metric_from_request("accounts.active.disable:4567|c|@1.0")
        )
        self.assertEqual(
            Counter("float_rate", 345, 0.2),
            parse_metric_from_request("float_rate:345|c|@0.2")
        )

    def test_parse_counter_metric_from_invalid_request_raises_error(self):
        self.assertRaises(ValueError, parse_metric_from_request, "")
        self.assertRaises(
            ValueError,
            parse_metric_from_request,
            "invalid_request"
        )
        self.assertRaises(
            ValueError,
            parse_metric_from_request,
            "invalid request"
        )
        self.assertRaises(
            ValueError,
            parse_metric_from_request,
            "bad_value_separator|2|c"
        )
        self.assertRaises(
            ValueError,
            parse_metric_from_request,
            "missing_value:"
        )
        self.assertRaises(
            ValueError,
            parse_metric_from_request,
            "no type:2"
        )
        self.assertRaises(
            ValueError, parse_metric_from_request, "missing_type:2|")
        self.assertRaises(
            ValueError, parse_metric_from_request, "invalid_type:2|X")
        self.assertRaises(
            ValueError, parse_metric_from_request, "invalid_rate:2|c@hello")
        self.assertRaises(
            ValueError, parse_metric_from_request, "float_counter:2.2|c@1")

    def test_parse_timer_metric_from_request(self):
        self.assertEqual(
            Timer("exact", 1),
            parse_metric_from_request("exact:1|ms")
        )
        self.assertEqual(
            Timer("db query?", 2.4),
            parse_metric_from_request("db query?:2.4|ms")
        )
        self.assertEqual(
            Timer("db.query.with.rate", 10, 1),
            parse_metric_from_request("db.query.with.rate:10|ms|@1")
        )
        self.assertEqual(
            Timer("db_query_float_rate", 23.5, 0.5),
            parse_metric_from_request("db_query_float_rate:23.5|ms|@0.5")
        )

    def test_parse_timer_metric_from_invalid_request_raises_value_error(self):
        self.assertRaises(
            ValueError, parse_metric_from_request, "invalid_value:hello|ms")
        self.assertRaises(
            ValueError,
            parse_metric_from_request,
            "invalid_value:hello|ms|@0.5"
        )
        self.assertRaises(
            ValueError, parse_metric_from_request, "invalid_type:4|ms_")
        self.assertRaises(
            ValueError, parse_metric_from_request, "invalid_rate:4.2|ms@_")

    def test_parse_gauge_metric_from_request(self):
        self.assertEqual(
            Gauge("cpu_usage", 45.3),
            parse_metric_from_request("cpu_usage:45.3|g")
        )
        self.assertEqual(
            Gauge("mem usage?", 10240, 1),
            parse_metric_from_request("mem usage?:10240|g|@1")
        )
        self.assertEqual(
            Gauge("weird.gauge.with.rate", 23.3, 0.5),
            parse_metric_from_request("weird.gauge.with.rate:23.3|g|@0.5")
        )

    def test_parse_gauge_metric_from_invalid_request_raises_value_error(self):
        self.assertRaises(
            ValueError, parse_metric_from_request, "invalid_value:hello|g")
        self.assertRaises(
            ValueError,
            parse_metric_from_request,
            "invalid_value:hello|g|@0.5"
        )
        self.assertRaises(
            ValueError, parse_metric_from_request, "invalid_type:4|g_")
        self.assertRaises(
            ValueError, parse_metric_from_request, "invalid_rate:2.8|g@_")

    def test_parse_set_metric_from_request(self):
        self.assertEqual(
            Set("host", '127.0.0.1'),
            parse_metric_from_request("host:127.0.0.1|s")
        )
        self.assertEqual(
            Set("user id?", '12345', 1),
            parse_metric_from_request("user id?:12345|s|@1")
        )
        self.assertEqual(
            Set("weird.set.with.rate", '25.7', 0.5),
            parse_metric_from_request("weird.set.with.rate:25.7|s|@0.5")
        )

    def test_parse_set_metric_from_invalid_request_raises_value_error(self):
        self.assertRaises(
            ValueError, parse_metric_from_request, "invalid_type:4|s_")
        self.assertRaises(
            ValueError, parse_metric_from_request, "invalid_rate:name|s@_")

    def test_parse_gauge_delta_metric_from_request(self):
        self.assertEqual(
            GaugeDelta("cpu_usage", 6),
            parse_metric_from_request("cpu_usage:+6|g")
        )
        self.assertEqual(
            GaugeDelta("mem usage?", -2048, 1),
            parse_metric_from_request("mem usage?:-2048|g|@1")
        )
        self.assertEqual(
            GaugeDelta("weird.gauge.delta.with.rate", 23.3, 0.5),
            parse_metric_from_request(
                "weird.gauge.delta.with.rate:+23.3|g|@0.5")
        )

    def test_parse_gauge_delta_metric_from_invalid_request_raises_error(self):
        self.assertRaises(
            ValueError, parse_metric_from_request, "invalid_value:+hello|g")
        self.assertRaises(
            ValueError,
            parse_metric_from_request,
            "invalid_value:-hello|g|@0.5"
        )
        self.assertRaises(
            ValueError, parse_metric_from_request, "invalid_type:+4|g_")
        self.assertRaises(
            ValueError, parse_metric_from_request, "invalid_rate:-2.8|g@_")


class TestCounter(unittest.TestCase):
    def test_counter_constructor(self):
        counter = Counter('test', 5, 0.2)
        self.assertEqual(counter.name, 'test')
        self.assertEqual(counter.count, 5)
        self.assertEqual(counter.sample_rate, 0.2)

        counter_negative = Counter('negative', -10)
        self.assertEqual(counter_negative.count, -10)

    def test_metric_requires_a_non_empty_string_name(self):
        self.assertRaises(AssertionError, Counter, 0)
        self.assertRaises(AssertionError, Counter, '')

    def test_counter_default_count_is_zero(self):
        counter = Counter('test')
        self.assertEqual(counter.count, 0)

    def test_counter_default_sample_rate_is_one(self):
        counter = Counter('test')
        self.assertEqual(counter.sample_rate, 1)

    def test_count_should_be_integer(self):
        self.assertRaises(AssertionError, Counter, 'test', 1.2)
        counter = Counter('ok')

        def set_string_as_count():
            counter.count = 'not integer'

        self.assertRaises(AssertionError, set_string_as_count)
        counter.count = 2
        self.assertEqual(counter.count, 2)

    def test_sample_rate_should_be_numeric(self):
        self.assertRaises(
            AssertionError, Counter, 'string_sample_rate', 1, 'what?')
        counter = Counter('ok')
        counter.sample_rate = 0.4
        self.assertEqual(counter.sample_rate, 0.4)
        counter.sample_rate = 2
        self.assertEqual(counter.sample_rate, 2)

    def test_sample_rate_should_be_positive(self):
        self.assertRaises(AssertionError, Counter, 'negative', 1, -2.3)
        self.assertRaises(AssertionError, Counter, 'zero', 1, 0)

    def test_to_request(self):
        counter = Counter('something')
        self.assertEqual(counter.to_request(), 'something:0|c')

        counter2 = Counter('another', 3)
        self.assertEqual(counter2.to_request(), 'another:3|c')

        counter3 = Counter('again', -2, 0.7)
        self.assertEqual(counter3.to_request(), 'again:-2|c|@0.7')

    def test_equality_check(self):
        counter = Counter('counter1')
        with self.assertRaises(AssertionError):
            counter != 'I am not a Counter'

        different_name = Counter('counter2')
        self.assertTrue(counter != different_name)

        different_count = Counter('counter1', 10)
        self.assertTrue(counter != different_count)

        different_rate = Counter('counter1', 1, 0.5)
        self.assertTrue(counter != different_rate)

        same = Counter('counter1')
        self.assertTrue(same == counter)


class TestTimer(unittest.TestCase):
    def test_constructor(self):
        timer = Timer('test', 5.1, 0.2)
        self.assertEqual(timer.name, 'test')
        self.assertEqual(timer.milliseconds, 5.1)
        self.assertEqual(timer.sample_rate, 0.2)

    def test_metric_requires_a_non_empty_string_name(self):
        self.assertRaises(AssertionError, Timer, 0, 0.1)
        self.assertRaises(AssertionError, Timer, '', 0.1)

    def test_default_sample_rate_is_one(self):
        timer = Timer('test', 0.1)
        self.assertEqual(timer.sample_rate, 1)

    def test_millisecond_should_be_numeric(self):
        self.assertRaises(AssertionError, Timer, 'test', '')
        timer = Timer('ok', 0.3)
        self.assertEqual(timer.milliseconds, 0.3)
        timer.milliseconds = 2
        self.assertEqual(timer.milliseconds, 2)

    def test_millisecond_should_not_be_negative(self):
        self.assertRaises(AssertionError, Timer, 'test', -4.2)
        timer = Timer('ok', 0.0)
        self.assertEqual(timer.milliseconds, 0.0)

    def test_sample_rate_should_be_numeric(self):
        self.assertRaises(
            AssertionError, Timer, 'string_sample_rate', 1.0, 's')
        timer = Timer('ok', 0.1)
        timer.sample_rate = 0.3
        self.assertEqual(timer.sample_rate, 0.3)
        timer.sample_rate = 2
        self.assertEqual(timer.sample_rate, 2)

    def test_sample_rate_should_be_positive(self):
        self.assertRaises(AssertionError, Timer, 'negative', 1.2, -4.0)
        self.assertRaises(AssertionError, Timer, 'zero', 1.2, 0)

    def test_to_request(self):
        timer = Timer('ok', 0.2)
        self.assertEqual(timer.to_request(), 'ok:0.2|ms')

        timer2 = Timer('another', 45.2)
        self.assertEqual(timer2.to_request(), 'another:45.2|ms')

        timer3 = Timer('again', 12.3, 0.8)
        self.assertEqual(timer3.to_request(), 'again:12.3|ms|@0.8')

    def test_equality_check(self):
        timer = Timer('timer1', 10)
        with self.assertRaises(AssertionError):
            timer != 'I am not a Timer'

        different_name = Timer('timer2', 10)
        self.assertTrue(timer != different_name)

        different_time = Timer('timer1', 25)
        self.assertTrue(timer != different_time)

        different_rate = Timer('timer1', 10, 0.5)
        self.assertTrue(timer != different_rate)

        same = Timer('timer1', 10)
        self.assertTrue(same == timer)


class TestGauge(unittest.TestCase):
    def test_constructor(self):
        gauge = Gauge('test', 5, 0.2)
        self.assertEqual(gauge.name, 'test')
        self.assertEqual(gauge.value, 5)
        self.assertEqual(gauge.sample_rate, 0.2)

    def test_metric_requires_a_non_empty_string_name(self):
        self.assertRaises(AssertionError, Gauge, 0, 1)
        self.assertRaises(AssertionError, Gauge, '', 2)

    def test_default_sample_rate_is_one(self):
        gauge = Gauge('test', 3)
        self.assertEqual(gauge.sample_rate, 1)

    def test_value_should_be_numeric(self):
        self.assertRaises(AssertionError, Gauge, 'string_val', '')
        gauge = Gauge('ok', 0.3)

        def set_value_as_string():
            gauge.value = 'not float'

        self.assertRaises(AssertionError, set_value_as_string)
        gauge.value = 2.0
        self.assertEqual(gauge.value, 2.0)
        gauge.value = 63
        self.assertEqual(gauge.value, 63)

    def test_value_should_not_be_negative(self):
        self.assertRaises(AssertionError, Gauge, 'test', -2)
        gauge = Gauge('ok', 0)

        def set_negative_value():
            gauge.value = -4.5

        self.assertRaises(AssertionError, set_negative_value)

    def test_sample_rate_should_be_numeric(self):
        self.assertRaises(
            AssertionError, Gauge, 'string_sample_rate', 1.0, 's')
        gauge = Gauge('ok', 4)
        gauge.sample_rate = 0.3
        self.assertEqual(gauge.sample_rate, 0.3)
        gauge.sample_rate = 2
        self.assertEqual(gauge.sample_rate, 2)

    def test_sample_rate_should_be_positive(self):
        self.assertRaises(AssertionError, Gauge, 'negative', 10, -4.0)
        self.assertRaises(AssertionError, Gauge, 'zero', 10, 0)

    def test_to_request(self):
        gauge = Gauge('ok', 0.2)
        self.assertEqual(gauge.to_request(), 'ok:0.2|g')

        gauge2 = Gauge('another', 237)
        self.assertEqual(gauge2.to_request(), 'another:237|g')

        gauge3 = Gauge('again', 11.8, 0.4)
        self.assertEqual(gauge3.to_request(), 'again:11.8|g|@0.4')

    def test_equality_check(self):
        gauge = Gauge('cpu', 10)
        with self.assertRaises(AssertionError):
            gauge != 'I am not a Gauge'

        different_name = Gauge('memory', 10)
        self.assertTrue(gauge != different_name)

        different_value = Gauge('cpu', 25)
        self.assertTrue(gauge != different_value)

        different_rate = Gauge('cpu', 10, 0.5)
        self.assertTrue(gauge != different_rate)

        same = Gauge('cpu', 10)
        self.assertTrue(same == gauge)


class TestSet(unittest.TestCase):
    def test_constructor(self):
        set_ = Set('unique', 5)
        self.assertEqual(set_.name, 'unique')
        self.assertEqual(set_.value, 5)

    def test_metric_requires_a_non_empty_string_name(self):
        self.assertRaises(AssertionError, Set, 0, 1)
        self.assertRaises(AssertionError, Set, '', 2)

    def test_value_should_be_hashable(self):
        self.assertRaises(AssertionError, Set, 'not_hashable', [])
        set_ = Set('ok', 4)
        set_.value = 2.0
        self.assertEqual(set_.value, 2.0)
        set_.value = 'something hashable'
        self.assertEqual(set_.value, 'something hashable')

    def test_equality_check(self):
        metric = Set('uid', 5)
        with self.assertRaises(AssertionError):
            metric != 'I am not a Set'

        different_name = Set('userid', 5)
        self.assertTrue(metric != different_name)

        different_value = Set('uid', 25)
        self.assertTrue(metric != different_value)

        different_rate = Set('uid', 10, 0.5)
        self.assertTrue(metric != different_rate)

        same = Set('uid', 5)
        self.assertTrue(same == metric)


class TestGaugeDelta(unittest.TestCase):
    def test_constructor(self):
        gauge_delta = GaugeDelta('unique', 5)
        self.assertEqual(gauge_delta.name, 'unique')
        self.assertEqual(gauge_delta.delta, 5)

    def test_delta_should_be_numeric(self):
        self.assertRaises(AssertionError, GaugeDelta, 'string_val', '')
        gauge_delta = GaugeDelta('ok', 0.3)
        gauge_delta.delta = 2.0
        self.assertEqual(gauge_delta.delta, 2.0)
        gauge_delta.delta = 27
        self.assertEqual(gauge_delta.delta, 27)

    def test_to_request(self):
        gauge_delta = GaugeDelta('ok', 0.2)
        self.assertEqual(gauge_delta.to_request(), 'ok:+0.2|g')

        gauge_delta2 = GaugeDelta('another', -43)
        self.assertEqual(gauge_delta2.to_request(), 'another:-43|g')

        gauge_delta3 = GaugeDelta('again', 15, 0.4)
        self.assertEqual(gauge_delta3.to_request(), 'again:+15|g|@0.4')

    def test_equality_check(self):
        metric = GaugeDelta('cpu', 5)
        with self.assertRaises(AssertionError):
            metric != 'I am not a GaugeDelta'

        different_name = GaugeDelta('memory', 5)
        self.assertTrue(metric != different_name)

        different_value = GaugeDelta('cpu', 25)
        self.assertTrue(metric != different_value)

        different_rate = GaugeDelta('cpu', 5, 0.5)
        self.assertTrue(metric != different_rate)

        same = GaugeDelta('cpu', 5)
        self.assertTrue(same == metric)

if __name__ == '__main__':
    unittest.main()
